
__revision__ = "$Id: sml_parser.py,v 1.11 2001/07/28 22:08:04 nascheme Exp $"

from __future__ import generators, nested_scopes
import re, sys
from text_chunk import TextChunk as Context

VALID_TAGS = ("B", "EM", "I", "PL", "S", "TT", "U",
              "0", "1", "2", "3", "4", "5", "6", "7", "8", "9",
              "r", "g", "b", "c", "m", "y", "k", "w")

BOOLEAN_TAGS = {'B':'set_bold',
                'I':'set_italics',
                'TT':'set_tt'
                }
    
class ParseError(Exception):
    pass


def parse (file):
    """Parse input coming from the file-like object 'file'.  Generate a
    stream of TextChunks.
    """
    
    tag_re = re.compile("""
    < (/?) ([A-Za-z0-9]+) > 
    """, re.VERBOSE)

    BLOCK_SIZE = 1024
    tag_stack = []
    context_stack = [Context()]

    def extend_context_stack ():
        new = context_stack[-1].copy()
        context_stack.append(new)
        return new

    def update_context (context, tag_name):
        if BOOLEAN_TAGS.has_key(tag_name):
            meth = getattr(context, BOOLEAN_TAGS[tag_name])
            meth(1)
        elif tag_name == 'S':
            # Set the s flag and unset the emph flag
            context.set_strong(1)
            context.set_emph(0)
        elif tag_name == 'EM':
            # If already inside an <S>, <EM> is irrelevant, so ignore it.
            if not context.strong:
                context.set_emph(not context.emph)
        elif tag_name == 'PL':
            context.set_under(0)
            context.set_bold(0)
            context.set_emph(0)
            context.set_italics(0)
            context.set_strong(0)
            context.set_tt(0)
        elif tag_name == 'U':
            if context.under < 3:
                context.under += 1

        elif len(tag_name) == 1 and tag_name in '0123456789':
            context.set_size(int(tag_name))

        elif len(tag_name) == 1 and tag_name in 'rgbcmykw':
            context.set_color(tag_name)

        else:
            raise RuntimeError, "Unknown tag %r (can't happen!)" % tag_name

    def create_chunk(text):
        chunk = context_stack[-1].copy()
        chunk.set_text(text)
        return chunk
        
        
    block = file.read(BLOCK_SIZE)

    while 1:

        
        # Check for tag
        m = tag_re.search(block)
        if m:
            text = block[:m.start()]
            block = block[m.end():]
            closing = m.group(1) == "/"
            tag_name = m.group(2)

            if text:
                yield create_chunk(text)

            # Check for a valid tag name
            if tag_name not in VALID_TAGS:
                raise ParseError, "Invalid tag: %s" % repr(tag_name)

            if closing:
                if not tag_stack:
                    raise ParseError(
                        "Closing tag %r before any start tags" %  tag_name)

                if tag_name != tag_stack[-1]:
                    raise ParseError("Mismatched tags: closing tag %r doesn't "
                                     "match opening %r" % (tag_name,
                                                           tag_stack[-1]))

                tag_stack.pop()
                context_stack.pop()

            else:
                # Opening tag
                tag_stack.append(tag_name)
                context = extend_context_stack()
                update_context(context, tag_name)

        else:
            # no tags, keep reading
            more = file.read(BLOCK_SIZE)
            if not more:
                break
            block += more

    if block:
        yield create_chunk(block)

