from __future__ import generators


def lift_simple_tag (tree, tag):
    """Move simple (tags the affect attributes that are independent of other
    attributes, like B) nodes as high in the tree as possible."""
    for child in tree.children:
        if child.children:
            lift_simple_tag(child, tag)
    if (len(tree.children) == 1 and tree.tag != tag and
            not (tree.tag == "EM" and tag == "S")):
        child = tree.children[0]
        if child.tag == tag:
            (tree.tag, child.tag) = (child.tag, tree.tag)
        
def merge_simple(tree):
    """Try merge two peer nodes in the tree into one (in a simple minded
    way).  Return the number of merges done.
    """

    # "<I><B>a</B></I><I>b</I>" -> "<I><B>a</B>b</I>"

    merges = 0
    for child in tree.children:
        if child.children:
            merges += merge_simple(child)
    i = 0
    while i < len(tree.children)-1:
        a = tree.children[i]
        b = tree.children[i+1]
        if a.tag and a.tag == b.tag:
            a.children.extend(b.children) # add b's children to a
            del tree.children[i+1] # remove b
            merges += 1
        i += 1
    return merges
    

def merge_class (tree, in_class):

    # <r>foo</r><g>bar</g><r>baz</r> -> <r>foo<g>bar</g>baz</r>
    # <0>foo</0><1>bar</1><0>baz</0> -> <0>foo<1>bar</1>baz</0>

    merges = 0
    for child in tree.children:
        if child.children:
            merges += merge_class(child, in_class)
    i = 0
    while i < len(tree.children):
        child = tree.children[i]
        if in_class(child.tag):
            j = i+1
            while j < len(tree.children):
                child2 = tree.children[j]
                if not in_class(child2.tag):
                    break
                if child2.tag == child.tag:
                    nodes = tree.children[i:j+1]
                    nodes[:1] = child.children
                    nodes[-1:] = child2.children
                    child.children = nodes
                    tree.children[i:j+1] = [child]
                    merges += 1
                    break
                j += 1
        i += 1
    return merges


def lift_class (tree, in_class):
    for child in tree.children:
        if child.children:
            lift_class(child, in_class)
    if len(tree.children) == 1 and not in_class(tree.tag):
        child = tree.children[0]
        if in_class(child.tag):
            (tree.tag, child.tag) = (child.tag, tree.tag)


COLOR_TAGS = { "r" : 1, "g" : 1, "b" : 1, "c" : 1, "m" : 1, "y" : 1,
               "k" : 1, "w" : 1 }
is_color = COLOR_TAGS.has_key

def lift_color (tree):
    lift_class(tree, is_color)

def merge_color (tree):
    return merge_class(tree, is_color)


SIZE_TAGS = { "0" : 1, "1" : 1, "2" : 1, "3" : 1, "4" : 1, "5" : 1,
               "6" : 1, "7" : 1, "8" : 1, "9" : 1 }
is_size = SIZE_TAGS.has_key

def lift_size (tree):
    lift_class(tree, is_size)

def merge_size (tree):
    return merge_class(tree, is_size)
